"""Defines aspects for Items.

Aspects form intermediate items between tools and items.
"""

from gaphas.aspect.connector import ConnectionSink, Connector
from gaphas.aspect.handlemove import HandleMove, item_at_point
from gaphas.aspect.move import Move
